//=============================================================================
// ColorGeometryDemo.cpp by Kyoung Shin Park (Spring 2010)
//
// Demonstrates how to create the geometry of a colored cube and a colored pyramid
//
// Controls: Use mouse to orbit and zoom; use the 'W' and 'S' keys to 
//           alter the height of the camera.
//=============================================================================


#include "d3dApp.h"
#include "DirectInput.h"
#include <crtdbg.h>
#include "GfxStats.h"
#include <list>
#include "Vertex.h"

const D3DCOLOR WHITE   = D3DCOLOR_XRGB(255, 255, 255); // 0xffffffff
const D3DCOLOR BLACK   = D3DCOLOR_XRGB(0, 0, 0);       // 0xff000000
const D3DCOLOR RED     = D3DCOLOR_XRGB(255, 0, 0);     // 0xffff0000
const D3DCOLOR GREEN   = D3DCOLOR_XRGB(0, 255, 0);     // 0xff00ff00
const D3DCOLOR BLUE    = D3DCOLOR_XRGB(0, 0, 255);     // 0xff0000ff
const D3DCOLOR YELLOW  = D3DCOLOR_XRGB(255, 255, 0);   // 0xffffff00
const D3DCOLOR CYAN    = D3DCOLOR_XRGB(0, 255, 255);   // 0xff00ffff
const D3DCOLOR MAGENTA = D3DCOLOR_XRGB(255, 0, 255);   // 0xffff00ff

const VertexCol CubeVertices[8] =
{
	{-1.0f, -1.0f, -1.0f, WHITE},
	{-1.0f,  1.0f, -1.0f, BLACK},
	{ 1.0f,  1.0f, -1.0f, RED},
	{ 1.0f, -1.0f, -1.0f, GREEN},
	{-1.0f, -1.0f,  1.0f, BLUE},
	{-1.0f,  1.0f,  1.0f, YELLOW},
	{ 1.0f,  1.0f,  1.0f, CYAN},
	{ 1.0f, -1.0f,  1.0f, MAGENTA}
};

const WORD CubeIndices[36] =
{
	// front side
	0, 1, 2,
	0, 2, 3,

	// back side
	4, 6, 5,
	4, 7, 6,

	// left side
	4, 5, 1,
	4, 1, 0,

	// right side
	3, 2, 6,
	3, 6, 7,

	// top
	1, 5, 6,
	1, 6, 2,

	// bottom
	4, 0, 3,
	4, 3, 7
};

const VertexCol PyramidVertices[5] =
{
	{3.0f,  0.5f,  0.0f, BLACK},
	{2.5f, -0.5f, -1.0f, RED},
	{2.5f, -0.5f,  1.0f, RED},
	{3.5f, -0.5f,  1.0f, BLUE},
	{3.5f, -0.5f, -1.0f, BLUE}
};

const WORD PyramidIndices[18] =
{
	// front & side
	0, 1, 2,
	0, 2, 3,
	0, 3, 4,
	0, 4, 1,
	// bottom
	1, 3, 2,
	1, 4, 3
};

class ColorGeometryDemo : public D3DApp
{
public:
	ColorGeometryDemo(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP);
	~ColorGeometryDemo();

	bool checkDeviceCaps();
	void onLostDevice();
	void onResetDevice();
	void updateScene(float dt);
	void drawScene();

	// Helper methods
	void buildVertexBuffer();
	void buildIndexBuffer();
	void buildViewMtx();
	void buildProjMtx();

private:
	GfxStats* mGfxStats;
	
	IDirect3DVertexBuffer9* m_CubeVB;
	IDirect3DIndexBuffer9*  m_CubeIB;
	IDirect3DVertexBuffer9* m_PyramidVB;
	IDirect3DIndexBuffer9*  m_PyramidIB;

	float mCameraRotationY;
	float mCameraRadius;
	float mCameraHeight;

	D3DXMATRIX mView;
	D3DXMATRIX mProj;
};


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE prevInstance,
				   PSTR cmdLine, int showCmd)
{
	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
		_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	ColorGeometryDemo app(hInstance, "Color Geometry Demo", D3DDEVTYPE_HAL, D3DCREATE_HARDWARE_VERTEXPROCESSING);
	gd3dApp = &app;

	DirectInput di(DISCL_NONEXCLUSIVE|DISCL_FOREGROUND, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
	gDInput = &di;

	return gd3dApp->run();
}

ColorGeometryDemo::ColorGeometryDemo(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP)
: D3DApp(hInstance, winCaption, devType, requestedVP)
{
	if(!checkDeviceCaps())
	{
		MessageBox(0, "checkDeviceCaps() Failed", 0, 0);
		PostQuitMessage(0);
	}

	mGfxStats = new GfxStats();

	mCameraRadius    = 10.0f;
	mCameraRotationY = 1.2 * D3DX_PI;
	mCameraHeight    = 5.0f;

	buildVertexBuffer();
	buildIndexBuffer();

	onResetDevice();

	InitAllVertexDeclarations();
}

ColorGeometryDemo::~ColorGeometryDemo()
{
	delete mGfxStats;
	ReleaseCOM(m_CubeVB);
	ReleaseCOM(m_CubeIB);
	ReleaseCOM(m_PyramidVB);
	ReleaseCOM(m_PyramidIB);

	DestroyAllVertexDeclarations();
}

bool ColorGeometryDemo::checkDeviceCaps()
{
	return true;
}

void ColorGeometryDemo::onLostDevice()
{
	mGfxStats->onLostDevice();
}

void ColorGeometryDemo::onResetDevice()
{
	mGfxStats->onResetDevice();

	// The aspect ratio depends on the backbuffer dimensions, which can 
	// possibly change after a reset.  So rebuild the projection matrix.
	buildProjMtx();
}

void ColorGeometryDemo::updateScene(float dt)
{
	// One cube has 8 vertice and 12 triangles.
	mGfxStats->setVertexCount(8);
	mGfxStats->setTriCount(12);
	mGfxStats->update(dt);

	// Get snapshot of input devices.
	gDInput->poll();

	// Check input.
	if( gDInput->keyDown(DIK_W) )	 
		mCameraHeight   += 25.0f * dt;
	if( gDInput->keyDown(DIK_S) )	 
		mCameraHeight   -= 25.0f * dt;

	// Divide by 50 to make mouse less sensitive. 
	mCameraRotationY += gDInput->mouseDX() / 100.0f;
	mCameraRadius    += gDInput->mouseDY() / 25.0f;

	// If we rotate over 360 degrees, just roll back to 0
	if( fabsf(mCameraRotationY) >= 2.0f * D3DX_PI ) 
		mCameraRotationY = 0.0f;

	// Don't let radius get too small.
	if( mCameraRadius < 5.0f )
		mCameraRadius = 5.0f;

	// The camera position/orientation relative to world space can 
	// change every frame based on input, so we need to rebuild the
	// view matrix every frame with the latest changes.
	buildViewMtx();
}


void ColorGeometryDemo::drawScene()
{
	// Clear the backbuffer and depth buffer.
	HR(gd3dDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xffffffff, 1.0f, 0));
	HR(gd3dDevice->BeginScene());

	// World matrix is identity.
	D3DXMATRIX W;
	D3DXMatrixIdentity(&W);
	HR(gd3dDevice->SetTransform(D3DTS_WORLD, &W));
	HR(gd3dDevice->SetTransform(D3DTS_VIEW, &mView));
	HR(gd3dDevice->SetTransform(D3DTS_PROJECTION, &mProj));

	HR(gd3dDevice->SetRenderState(D3DRS_LIGHTING, false));
	HR(gd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID));
	//HR(gd3dDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT));

	// Let Direct3D know the vertex buffer, index buffer and vertex 
	// declaration we are using, and then draw primitives.
	// Cube
	HR(gd3dDevice->SetStreamSource(0, m_CubeVB, 0, sizeof(VertexCol)));
	HR(gd3dDevice->SetIndices(m_CubeIB));
	HR(gd3dDevice->SetVertexDeclaration(VertexCol::Decl));
	HR(gd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, 8, 0, 12));
	// Pyramid
	HR(gd3dDevice->SetStreamSource(0, m_PyramidVB, 0, sizeof(VertexCol)));
	HR(gd3dDevice->SetIndices(m_PyramidIB));
	HR(gd3dDevice->SetVertexDeclaration(VertexCol::Decl));
	HR(gd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, 5, 0, 6));
	
	mGfxStats->display();

	HR(gd3dDevice->EndScene());

	// Present the backbuffer.
	HR(gd3dDevice->Present(0, 0, 0, 0));
}

void ColorGeometryDemo::buildVertexBuffer()
{
	// Obtain a pointer to a new vertex buffer.
	HR(gd3dDevice->CreateVertexBuffer(8 * sizeof(VertexCol), D3DUSAGE_WRITEONLY,
		0, D3DPOOL_MANAGED, &m_CubeVB, 0));

	// Now lock it to obtain a pointer to its internal data, and write the
	// cube's vertex data.
	VertexCol* v = 0;
	HR(m_CubeVB->Lock(0, 0, (void**)&v, 0));
	for (int i = 0; i < 8; i++) {
		v[i] = VertexCol(CubeVertices[i]);
	}
	HR(m_CubeVB->Unlock());

	// Pyramid's vertex data.
	HR(gd3dDevice->CreateVertexBuffer(5 * sizeof(VertexCol), D3DUSAGE_WRITEONLY,
		0, D3DPOOL_MANAGED, &m_PyramidVB, 0));
	v = 0;
	HR(m_PyramidVB->Lock(0, 0, (void**)&v, 0));
	for (int i = 0; i < 5; i++) {
		v[i] = VertexCol(PyramidVertices[i]);
	}
	HR(m_PyramidVB->Unlock());
}

void ColorGeometryDemo::buildIndexBuffer()
{
	// Obtain a pointer to a new index buffer.
	HR(gd3dDevice->CreateIndexBuffer(36 * sizeof(WORD), D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_CubeIB, 0));

	// Now lock it to obtain a pointer to its internal data, and write the
	// cube's index data.
	WORD* k = 0;
	HR(m_CubeIB->Lock(0, sizeof(CubeIndices), (void**)&k, 0));
	memcpy(k, CubeIndices, sizeof(CubeIndices));
	HR(m_CubeIB->Unlock());

	// Pyramid's index data.
	HR(gd3dDevice->CreateIndexBuffer(18 * sizeof(WORD), D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_PyramidIB, 0));
	k = 0;
	HR(m_PyramidIB->Lock(0, sizeof(PyramidIndices), (void**)&k, 0));
	memcpy(k, PyramidIndices, sizeof(PyramidIndices));
	HR(m_PyramidIB->Unlock());
}

void ColorGeometryDemo::buildViewMtx()
{
	float x = mCameraRadius * cosf(mCameraRotationY);
	float z = mCameraRadius * sinf(mCameraRotationY);
	D3DXVECTOR3 pos(x, mCameraHeight, z);
	D3DXVECTOR3 target(0.0f, 0.0f, 0.0f);
	D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
	D3DXMatrixLookAtLH(&mView, &pos, &target, &up);
}

void ColorGeometryDemo::buildProjMtx()
{
	float w = (float)md3dPP.BackBufferWidth;
	float h = (float)md3dPP.BackBufferHeight;
	D3DXMatrixPerspectiveFovLH(&mProj, D3DX_PI * 0.25f, w/h, 1.0f, 5000.0f);
}